/*
  h_file.h   Nov98  JS  ---  please look at h_file.c for more infos
  
*/
#ifndef _H_FILE_H
#define _H_FILE_H

#include "config.h"
#include "defs.h"
//#include "hilbert.h" /* tbase */
// #include "array_m.h" /* tvec=scplx */

/* ----------    packed-struct-array (obsolete) -----------------------
  struct thxy {			      // y,h[x,y] if(x=y)x++;
    int  xy __attribute__ ((packed)) ; // 4 Byte 
    char rr __attribute__ ((packed)) ; // 1 Byte 
  }; // 5 Byte packet, 8 Byte aligned!
 * __sparc-gcc2.7.2.3:    cplx=20Byte buserror in nhm hxy.rr=r;
 * __sgi-CC: no __attribute__ 
 * __hp9000s800-gcc2.7.0: cplx=20Byte buserror in nhm hxy.rr=r;
 * __alpha_ev5__-gcc2.95.2: cplx=20Byte unaligned trap at ... (/proc/kmsg)
 * 2D-array (struct of char[3],...) and own assignment-operator to slow
 * struct thxy { char xy[S_int], rr[S_char] } + casting-operatoren = same problem
 */

/*
  if HRMAX greater than 0,
  the program stores a index to the double or complex value of the matrix
  element, else the matrix element itself is stored to the disk
 --- if an error apears, choose xy as a next larger type ---
 --- please edit HRMAX too !!! ---
 unsigned char , unsigned int , ULong
 */
#ifndef HRMAX		/* can be defined in config.h */
#define HRMAX   1023  	/* 255, 1024 or higher, if ERR(9100), v2.33 */
#endif
#if   HRMAX<256
#define HR_TYPE	unsigned char	/* char bis 256 */
#elif HRMAX<65536
#define HR_TYPE	unsigned short	/* short bis 65536 */
#else
#define HR_TYPE	unsigned int	/* int bis 256^4 */
#endif

/* y-index + next-x-flag of H element */
#define BY_TYPE unsigned short  /* char,short,int = lg2(blocks*mpi) v2.33 */
#ifndef XY_TYPE                 /* allows -DXY_TYPE=__int128 outside */
#define XY_TYPE unsigned int	/* Ulong, needes more memory + I/O! */
         /* 32bit is enough for 2G-elements to 8B = 16GB/MPItask */
         /* for 32GB/node no overflows with minimum 2 tasks/node */
#endif
/* XY_FLAG: bit in XY_TYPE = last element in H-lines, ToDo: better num-list */ 
#define XY_FLAG (((XY_TYPE)1)<<(sizeof(XY_TYPE)*8-1))

/* ToDo2015: IEEE 754-2008 16bit-Minifloats
 * ToDo2015: or 3bitparam*3bitn*0.5*sqrt(5bitn1/5bitn2)=16bit */
#if (!defined(HRMAX)) || (HRMAX<=0)
#define HR_TYPE scplx
#endif

/* AH NZX VS shbuf from spins.h to h_file.h(matrix-store-defs) 2017-01-23 */
/* ToDo: alloc in dynamic way and store max needed size!?
 *       better fix for MPI_transfers (but fill with mpi*n Hlines[blk]
 * NZXMAX max# of elements of one line in a HBlock, smaller is faster
 *   see output of h_nz_x_max for a optimum value
 * NUM_AHEAD_LINES is an buffer to store H lines ahead for less MPI calls
 * 2017-01-13 64 is to bad scaling see y2016/lrzGarching*.txt AH=32K HBLen=4M
 *          1024 used now (eats more memory, p.e. 8M growed to 41MB)
 *          see plot speed/core vs. cores (using min. nodes for growing n1)
 */
#ifndef  NUM_AHEAD_LINES  /* better cache usage, but bigger buffer size */
# define NUM_AHEAD_LINES (1024)   /* must be 1 or greater, static memory! */
#endif
/* maximum number of non-zero elements per matrix line
 *   keep it as minimum to minimize buffer size (NUM_AHEAD_LINES*NZXMAX) */
#ifndef NZXMAX  /* 2015-05-11 multiple of 32 for VS-vectors at AH=1 */
# ifdef  Nw
#  ifdef  AddSS
#    define NZXMAX ((((NN/2)*(NN/2)+Nw+2)+31) & ~31)
#  else
#    define NZXMAX (((Nw+2)+31) & ~31)
#  endif
# else
#  define NZXMAX 256
# endif
#endif

/* number of elements per hnz-read-buffer (HBLen*(XYsize+HRsize))
 * 2017-01-23 MPI-I/O is the bottleneck, dont care about CPU caches here
 *  HBLen = NUM_AHEAD_LINES * NZX  to have same MPI-packetsize in SH + i100
 *  ToDo:  but to big using NZXMAX? do malloc, remove HBLen, AH=2**n
 */
//#ifndef HBLen    /* make it small, if H is not stored (no buffer needed) */
//  16.02.17 HBLen now depends from AH*NZX, synchron nonequal SH blocks
//           old: asynchron equal SH blocks, but buggy complex MPI sync
#define HBLen ((NZXMAX)*(NUM_AHEAD_LINES)) /* 2017-01-23 optimized for mpi */ 
// #endif
/* structure to hold stored H-matrix elements
 * direct access to nonaligned members cause trap on IRIX ([].xy, [].r)
 *   therefore xy[],r[]
 * Compressed sparse row (CSR, CRS or Yale format) ofs by XY_FLAG
 * ToDo17 store data sorted to destination nodes!
 */
struct thxy {	/* y,h[x,y] if(x=y)x++; compact sparse matrix storage */
   /* ToDo: add xx (Hsize+30%) and sort to mpi-blocks (remove by),
            simpler mpi code (noXY_FLAG),
            faster mpi (const HBLen*XY_TYPE=mpi_pckt_size)?, +traffic/2
            xx,rr stored local, xy stored at remote node */
   /* extra data can be stored on first diag element xx and xy */
   /* XY_FLAG=1bit*NZX*AH z.B. 33bit*n1, so extra ofs[AH] (2B) eats less */
   /* 2017-01-24 xy_ofs[] added to replace XY_FLAG + HBLen by NZX*AH */
   // 2017-02 better extra xy_ofs[node_n1], step 2 struct_shbuf[HBLen] 
#if 0
#if HBLen < 65536 /* 2B=16bit-ushort or 4B=32bit-uint ofs[0..l-1]+len[l] */
   unsigned short xy_ofs[NUM_AHEAD_LINES+1];
#else
   unsigned int   xy_ofs[NUM_AHEAD_LINES+1];
#endif
#endif
   /* ToDo.Problem {xy[],rr[]} has fixed max size, bad to store 
    * {xy,rr}[] may waste Bytes for alignment, but flex. size, by as xy-bits?
    *    ToDo: warn bad nzx, redesign later? (always some copying needed?)
    *   by could be recomputed log2(mpi_n)-rnd-read in mpi_n*8B 
    *    mpi_n=100k 17*rnd-read on 800kB
    *   xy+by as (by << logmaxxy) + xy 
    *  ToDo17: [HBLen].{[by,]xy,rr} to simply cut stored elements at nzx
    * define switch-Macro:
    *    #define M1(aaa,rr,idx) aaa.nz[idx].rr
    *    b_hbuf[blk].rr[ b_bnz[blk] ]=...     ... = tbh->xy[ ii ]
    *    b_hbuf[blk].nz[ b_bnz[blk] ].rr=...  ... = tbh->nz[ ii ].xy
    */
#if 0
   XY_TYPE xy[HBLen]; /* column y of element h[] (4B=32bit idx + XY_FLAG) */
   BY_TYPE by[HBLen]; /* block y  of element h[] (2B=16bit mpi node) */
   HR_TYPE rr[HBLen]; /* [index to] NZ-Element (uchar,2B=short,int,scplx) */
#else
  struct {  /* 2017-02-20 */
    XY_TYPE xy __attribute__((__packed__)); /* column y of nzH[] (4B=32bit idx + XY_FLAG) */
    BY_TYPE by __attribute__((__packed__)); /* block y  of nzH[] (2B=16bit mpi node) */
    HR_TYPE rr __attribute__((__packed__)); /* [index to] nzH (uchar,2B=short,int,scplx) */
  } nz[HBLen];
#endif
   /* ToDo: better variant b or c ???
      variant a) all elements + flag EOL (or better xy_ofs)
      variant b) const number of nondiag elements (max hnzline)
      variant c) diag (has coded num nondiag) + nondiag
     used to write blocks to disk/storage but also should fit to CPU cache
    */
};

#define Hsize sizeof(struct thxy)

#ifndef B_NUM
#error "we need B_NUM here defined in block.h"
#endif

extern int h_error;   /* set on write-error */
extern int h_state[B_NUM];    /* 0=closed, 1=ropen, 2=wopen */  


#if HRMAX>0
 /* table for 16bit(HR_SIZE) index to Hamilton-entries, one per thread */
 extern int  hr_len[B_NUM];
 extern cplx hr[B_NUM][HRMAX];
#endif

int h_ropen(int block);
int h_wopen(int block);
int h_rclose(int block);
int h_wclose(int block);
struct thxy *h_read(struct thxy *thb, int len, int block);

/* store/read H-element, block, column=xy+Flag_next_line, value */
/* void h_put(int block, ULong  xx, ULong  xy, cplx  r); */
/* void h_get(int block, ULong *xx, ULong *xy, cplx *r); */
size_t my_fread( char *ptr, size_t size, size_t nmemb, int block);
size_t my_fwrite( void *ptr, size_t size, size_t nmemb, int block);
extern int h_state[B_NUM];    /* 0=closed, 1=ropen, 2=wopen */
void err_fulltable();

void ini_thxy(ULong n1);
/* if vvv outut info */
void h_checkpoint(int vvv);
void h_restore(int vvv);
#endif
